/*
 * Translate Prolog format CSP to ASP
 */
:- dynamic clause_count/1.

trans(InFile) :-
	open(InFile, read, In),
	repeat,
  	  read(In, X),
	  translate(X),
	X == end_of_file,
	close(In).

translate(end_of_file) :- !.
translate(X) :-
	tr(X, Ss, []),
	!,
	output(Ss).
translate(X) :-
	output([error(X)]).

output([]).
output([S|Ss]) :-
	write_canonical(S), write('.'), nl,
	output(Ss).

tr([int,V,Domain]) -->
	tr_int(V, Domain).
tr([bool,V]) -->
	tr_bool(V).
tr([relation,R,N,Body]) -->
	tr_relation(R, N, Body).
tr([objective,Target,V]) -->
	[objective(Target,V)].
tr(Clause) -->
	tr_clause(Clause).

/*
 * Integer Variable
 */
tr_int(V, Domain) -->
	{tr_list(Domain, range, Domain1)},
	[int(V, Domain1)].

/*
 * Boolean Variable
 */
tr_bool(V) -->
	[bool(V)].

/*
 * Relation
 */
tr_relation(R, N, [Type|Tuples]) -->
	[relation(R, N, Type, Tuples)].

/*
 * Clause
 */
tr_clause([or|Lits]) --> !,
	{clause_number(N)},
	[constraint(N)],
	tr_literals(N, Lits).
tr_clause(Lit) -->
	tr_clause([or,Lit]).

clause_number(N) :-
	(retract(clause_count(N0)) ; N0 = 0),
	!,
	N is N0 + 1,
	assert(clause_count(N)).

tr_literals(_, []) --> !.
tr_literals(N, [Lit|Lits]) -->
	{tr_literal(Lit, Z)},
	[constraint_elem(N, Z)],
	tr_literals(N, Lits).

tr_literal(Lit, Lit) :- atomic(Lit), !.
tr_literal([not, Lit], not(Z)) :- !,
	tr_literal(Lit, Z).
tr_literal([wsum, AXs, le, B], Z) :- !,
	tr_wsum_le(AXs, B, Z).
tr_literal([alldifferent, Xs], Z) :- !,
	Z = alldifferent(Xs).
tr_literal([weightedsum, AXs, Comp, B], Z) :- !,
	tr_list(AXs, mul, AXs1),
	Z = wsum(AXs1, Comp, B).
tr_literal([cumulative, Tasks, Limit], Z) :- !,
	tr_list(Tasks, task, Tasks1),
	Z = cumulative(Tasks1, Limit).
tr_literal([element, Index, Terms, Value], Z) :- !,
	Z = element(Index, Terms, Value).
tr_literal([djsjunctive, Tasks], Z) :- !,
	tr_list(Tasks, task, Tasks1),
	Z = disjunctive(Tasks1).
tr_literal([lex_less, Xs, Ys], Z) :- !,
	Z = lex_less(Xs, Ys).
tr_literal([lex_lesseq, Xs, Ys], Z) :- !,
	Z = lex_lesseq(Xs, Ys).
tr_literal([nvalue, Count, Xs], Z) :- !,
	Z = nvalue(Count, Xs).
tr_literal([global_cardinality, Xs, Cards], Z) :- !,
	tr_list(Cards, card, Cards1),
	Z = global_cardinality(Xs, Cards1).
tr_literal([global_cardinality_with_costs, Xs, Cards, Costs, Cost], Z) :- !,
	tr_list(Cards, card, Cards1),
	tr_list(Costs, cost, Costs1),
	Z = global_cardinality_with_costs(Xs, Cards1, Costs1, Cost).
tr_literal([count, X, Xs, Comp, Count], Z) :- !,
	Z = count(X, Xs, Comp, Count).
tr_literal([R|Xs], Z) :-
	Z = rel(R, Xs).

tr_wsum_le([[A,X]], B, Z) :- A > 0, !,
	C is integer(floor(B/A)),
	Z = var_le(X, C).
tr_wsum_le([[A,X]], B, Z) :- A < 0, !,
	C is integer(ceiling(B/A)-1),
	Z = not(var_le(X, C)).
tr_wsum_le(AXs, B, Z) :-
	tr_list(AXs, mul, AXs1),
	Z = wsum(AXs1, le, B).

tr_list([], _, []).
tr_list([X|Xs], F, [Z|Zs]) :-
	Z =.. [F|X],
	tr_list(Xs, F, Zs).
